@@ -78,18 +78,19 @@ fbIn (uint32_t x, uint8_t y)
  */
 
 static void
-fbCompositeOver_x888x8x8888 (pixman_op_t      op,
+fbCompositeOver_x888x8x8888 (pixman_implementation_t *imp,
+			     pixman_op_t      op,
 			     pixman_image_t * pSrc,
 			     pixman_image_t * pMask,
 			     pixman_image_t * pDst,
-			     int16_t      xSrc,
-			     int16_t      ySrc,
-			     int16_t      xMask,
-			     int16_t      yMask,
-			     int16_t      xDst,
-			     int16_t      yDst,
-			     uint16_t     width,
-			     uint16_t     height)
+			     int32_t      xSrc,
+			     int32_t      ySrc,
+			     int32_t      xMask,
+			     int32_t      yMask,
+			     int32_t      xDst,
+			     int32_t      yDst,
+			     int32_t     width,
+			     int32_t     height)
 {
     uint32_t	*src, *srcLine;
     uint32_t    *dst, *dstLine;
@@ -135,18 +136,19 @@ fbCompositeOver_x888x8x8888 (pixman_op_t      op,
 }
 
 static void
-fbCompositeSolidMaskIn_nx8x8 (pixman_op_t      op,
+fbCompositeSolidMaskIn_nx8x8 (pixman_implementation_t *imp,
+			      pixman_op_t      op,
 			      pixman_image_t    *iSrc,
 			      pixman_image_t    *iMask,
 			      pixman_image_t    *iDst,
-			      int16_t      xSrc,
-			      int16_t      ySrc,
-			      int16_t      xMask,
-			      int16_t      yMask,
-			      int16_t      xDst,
-			      int16_t      yDst,
-			      uint16_t     width,
-			      uint16_t     height)
+			      int32_t      xSrc,
+			      int32_t      ySrc,
+			      int32_t      xMask,
+			      int32_t      yMask,
+			      int32_t      xDst,
+			      int32_t      yDst,
+			      int32_t     width,
+			      int32_t     height)
 {
     uint32_t	src, srca;
     uint8_t	*dstLine, *dst, dstMask;
@@ -217,18 +219,19 @@ fbCompositeSolidMaskIn_nx8x8 (pixman_op_t      op,
 
 
 static void
-fbCompositeSrcIn_8x8 (pixman_op_t      op,
+fbCompositeSrcIn_8x8 (pixman_implementation_t *imp,
+		      pixman_op_t      op,
 		      pixman_image_t  *iSrc,
 		      pixman_image_t  *iMask,
 		      pixman_image_t  *iDst,
-		      int16_t          xSrc,
-		      int16_t          ySrc,
-		      int16_t          xMask,
-		      int16_t          yMask,
-		      int16_t          xDst,
-		      int16_t          yDst,
-		      uint16_t         width,
-		      uint16_t         height)
+		      int32_t          xSrc,
+		      int32_t          ySrc,
+		      int32_t          xMask,
+		      int32_t          yMask,
+		      int32_t          xDst,
+		      int32_t          yDst,
+		      int32_t         width,
+		      int32_t         height)
 {
     uint8_t	*dstLine, *dst;
     uint8_t	*srcLine, *src;
@@ -265,18 +268,19 @@ fbCompositeSrcIn_8x8 (pixman_op_t      op,
 }
 
 static void
-fbCompositeSolidMask_nx8x8888 (pixman_op_t      op,
+fbCompositeSolidMask_nx8x8888 (pixman_implementation_t *imp,
+			       pixman_op_t      op,
 			       pixman_image_t * pSrc,
 			       pixman_image_t * pMask,
 			       pixman_image_t * pDst,
-			       int16_t      xSrc,
-			       int16_t      ySrc,
-			       int16_t      xMask,
-			       int16_t      yMask,
-			       int16_t      xDst,
-			       int16_t      yDst,
-			       uint16_t     width,
-			       uint16_t     height)
+			       int32_t      xSrc,
+			       int32_t      ySrc,
+			       int32_t      xMask,
+			       int32_t      yMask,
+			       int32_t      xDst,
+			       int32_t      yDst,
+			       int32_t     width,
+			       int32_t     height)
 {
     uint32_t	 src, srca;
     uint32_t	*dstLine, *dst, d, dstMask;
@@ -323,18 +327,19 @@ fbCompositeSolidMask_nx8x8888 (pixman_op_t      op,
 }
 
 static void
-fbCompositeSolidMask_nx8888x8888C (pixman_op_t op,
+fbCompositeSolidMask_nx8888x8888C (pixman_implementation_t *imp,
+				   pixman_op_t op,
 				   pixman_image_t * pSrc,
 				   pixman_image_t * pMask,
 				   pixman_image_t * pDst,
-				   int16_t      xSrc,
-				   int16_t      ySrc,
-				   int16_t      xMask,
-				   int16_t      yMask,
-				   int16_t      xDst,
-				   int16_t      yDst,
-				   uint16_t     width,
-				   uint16_t     height)
+				   int32_t      xSrc,
+				   int32_t      ySrc,
+				   int32_t      xMask,
+				   int32_t      yMask,
+				   int32_t      xDst,
+				   int32_t      yDst,
+				   int32_t     width,
+				   int32_t     height)
 {
     uint32_t	src, srca;
     uint32_t	*dstLine, *dst, d, dstMask;
@@ -396,18 +401,19 @@ fbCompositeSolidMask_nx8888x8888C (pixman_op_t op,
 }
 
 static void
-fbCompositeSolidMask_nx8x0888 (pixman_op_t op,
+fbCompositeSolidMask_nx8x0888 (pixman_implementation_t *imp,
+			       pixman_op_t op,
 			       pixman_image_t * pSrc,
 			       pixman_image_t * pMask,
 			       pixman_image_t * pDst,
-			       int16_t      xSrc,
-			       int16_t      ySrc,
-			       int16_t      xMask,
-			       int16_t      yMask,
-			       int16_t      xDst,
-			       int16_t      yDst,
-			       uint16_t     width,
-			       uint16_t     height)
+			       int32_t      xSrc,
+			       int32_t      ySrc,
+			       int32_t      xMask,
+			       int32_t      yMask,
+			       int32_t      xDst,
+			       int32_t      yDst,
+			       int32_t     width,
+			       int32_t     height)
 {
     uint32_t	src, srca;
     uint8_t	*dstLine, *dst;
@@ -458,18 +464,19 @@ fbCompositeSolidMask_nx8x0888 (pixman_op_t op,
 }
 
 static void
-fbCompositeSolidMask_nx8x0565 (pixman_op_t op,
+fbCompositeSolidMask_nx8x0565 (pixman_implementation_t *imp,
+			       pixman_op_t op,
 				  pixman_image_t * pSrc,
 				  pixman_image_t * pMask,
 				  pixman_image_t * pDst,
-				  int16_t      xSrc,
-				  int16_t      ySrc,
-				  int16_t      xMask,
-				  int16_t      yMask,
-				  int16_t      xDst,
-				  int16_t      yDst,
-				  uint16_t     width,
-				  uint16_t     height)
+				  int32_t      xSrc,
+				  int32_t      ySrc,
+				  int32_t      xMask,
+				  int32_t      yMask,
+				  int32_t      xDst,
+				  int32_t      yDst,
+				  int32_t     width,
+				  int32_t     height)
 {
     uint32_t	src, srca;
     uint16_t	*dstLine, *dst;
@@ -521,18 +528,19 @@ fbCompositeSolidMask_nx8x0565 (pixman_op_t op,
 }
 
 static void
-fbCompositeSolidMask_nx8888x0565C (pixman_op_t op,
+fbCompositeSolidMask_nx8888x0565C (pixman_implementation_t *imp,
+				   pixman_op_t op,
 				   pixman_image_t * pSrc,
 				   pixman_image_t * pMask,
 				   pixman_image_t * pDst,
-				   int16_t      xSrc,
-				   int16_t      ySrc,
-				   int16_t      xMask,
-				   int16_t      yMask,
-				   int16_t      xDst,
-				   int16_t      yDst,
-				   uint16_t     width,
-				   uint16_t     height)
+				   int32_t      xSrc,
+				   int32_t      ySrc,
+				   int32_t      xMask,
+				   int32_t      yMask,
+				   int32_t      xDst,
+				   int32_t      yDst,
+				   int32_t     width,
+				   int32_t     height)
 {
     uint32_t	src, srca;
     uint16_t	src16;
@@ -594,18 +602,19 @@ fbCompositeSolidMask_nx8888x0565C (pixman_op_t op,
 }
 
 static void
-fbCompositeSrc_8888x8888 (pixman_op_t op,
+fbCompositeSrc_8888x8888 (pixman_implementation_t *imp,
+			  pixman_op_t op,
 			 pixman_image_t * pSrc,
 			 pixman_image_t * pMask,
 			 pixman_image_t * pDst,
-			 int16_t      xSrc,
-			 int16_t      ySrc,
-			 int16_t      xMask,
-			 int16_t      yMask,
-			 int16_t      xDst,
-			 int16_t      yDst,
-			 uint16_t     width,
-			 uint16_t     height)
+			 int32_t      xSrc,
+			 int32_t      ySrc,
+			 int32_t      xMask,
+			 int32_t      yMask,
+			 int32_t      xDst,
+			 int32_t      yDst,
+			 int32_t     width,
+			 int32_t     height)
 {
     uint32_t	*dstLine, *dst, dstMask;
     uint32_t	*srcLine, *src, s;
@@ -640,18 +649,19 @@ fbCompositeSrc_8888x8888 (pixman_op_t op,
 }
 
 static void
-fbCompositeSrc_8888x0888 (pixman_op_t op,
+fbCompositeSrc_8888x0888 (pixman_implementation_t *imp,
+			  pixman_op_t op,
 			 pixman_image_t * pSrc,
 			 pixman_image_t * pMask,
 			 pixman_image_t * pDst,
-			 int16_t      xSrc,
-			 int16_t      ySrc,
-			 int16_t      xMask,
-			 int16_t      yMask,
-			 int16_t      xDst,
-			 int16_t      yDst,
-			 uint16_t     width,
-			 uint16_t     height)
+			 int32_t      xSrc,
+			 int32_t      ySrc,
+			 int32_t      xMask,
+			 int32_t      yMask,
+			 int32_t      xDst,
+			 int32_t      yDst,
+			 int32_t     width,
+			 int32_t     height)
 {
     uint8_t	*dstLine, *dst;
     uint32_t	d;
@@ -689,18 +699,19 @@ fbCompositeSrc_8888x0888 (pixman_op_t op,
 }
 
 static void
-fbCompositeSrc_8888x0565 (pixman_op_t op,
+fbCompositeSrc_8888x0565 (pixman_implementation_t *imp,
+			  pixman_op_t op,
 			 pixman_image_t * pSrc,
 			 pixman_image_t * pMask,
 			 pixman_image_t * pDst,
-			 int16_t      xSrc,
-			 int16_t      ySrc,
-			 int16_t      xMask,
-			 int16_t      yMask,
-			 int16_t      xDst,
-			 int16_t      yDst,
-			 uint16_t     width,
-			 uint16_t     height)
+			 int32_t      xSrc,
+			 int32_t      ySrc,
+			 int32_t      xMask,
+			 int32_t      yMask,
+			 int32_t      xDst,
+			 int32_t      yDst,
+			 int32_t     width,
+			 int32_t     height)
 {
     uint16_t	*dstLine, *dst;
     uint32_t	d;
@@ -741,18 +752,19 @@ fbCompositeSrc_8888x0565 (pixman_op_t op,
 }
 
 static void
-fbCompositeSrc_x888x0565 (pixman_op_t op,
+fbCompositeSrc_x888x0565 (pixman_implementation_t *imp,
+			  pixman_op_t op,
                           pixman_image_t * pSrc,
                           pixman_image_t * pMask,
                           pixman_image_t * pDst,
-                          int16_t      xSrc,
-                          int16_t      ySrc,
-                          int16_t      xMask,
-                          int16_t      yMask,
-                          int16_t      xDst,
-                          int16_t      yDst,
-                          uint16_t     width,
-                          uint16_t     height)
+                          int32_t      xSrc,
+                          int32_t      ySrc,
+                          int32_t      xMask,
+                          int32_t      yMask,
+                          int32_t      xDst,
+                          int32_t      yDst,
+                          int32_t     width,
+                          int32_t     height)
 {
     uint16_t	*dstLine, *dst;
     uint32_t	*srcLine, *src, s;
@@ -780,18 +792,19 @@ fbCompositeSrc_x888x0565 (pixman_op_t op,
 }
 
 static void
-fbCompositeSrcAdd_8000x8000 (pixman_op_t	op,
+fbCompositeSrcAdd_8000x8000 (pixman_implementation_t *imp,
+			     pixman_op_t	op,
 			     pixman_image_t * pSrc,
 			     pixman_image_t * pMask,
 			     pixman_image_t * pDst,
-			     int16_t      xSrc,
-			     int16_t      ySrc,
-			     int16_t      xMask,
-			     int16_t      yMask,
-			     int16_t      xDst,
-			     int16_t      yDst,
-			     uint16_t     width,
-			     uint16_t     height)
+			     int32_t      xSrc,
+			     int32_t      ySrc,
+			     int32_t      xMask,
+			     int32_t      yMask,
+			     int32_t      xDst,
+			     int32_t      yDst,
+			     int32_t     width,
+			     int32_t     height)
 {
     uint8_t	*dstLine, *dst;
     uint8_t	*srcLine, *src;
@@ -830,18 +843,19 @@ fbCompositeSrcAdd_8000x8000 (pixman_op_t	op,
 }
 
 static void
-fbCompositeSrcAdd_8888x8888 (pixman_op_t	op,
+fbCompositeSrcAdd_8888x8888 (pixman_implementation_t *imp,
+			     pixman_op_t	op,
 			     pixman_image_t * pSrc,
 			     pixman_image_t * pMask,
 			     pixman_image_t * pDst,
-			     int16_t      xSrc,
-			     int16_t      ySrc,
-			     int16_t      xMask,
-			     int16_t      yMask,
-			     int16_t      xDst,
-			     int16_t      yDst,
-			     uint16_t     width,
-			     uint16_t     height)
+			     int32_t      xSrc,
+			     int32_t      ySrc,
+			     int32_t      xMask,
+			     int32_t      yMask,
+			     int32_t      xDst,
+			     int32_t      yDst,
+			     int32_t     width,
+			     int32_t     height)
 {
     uint32_t	*dstLine, *dst;
     uint32_t	*srcLine, *src;
@@ -887,18 +901,19 @@ fbCompositeSrcAdd_8888x8888 (pixman_op_t	op,
 }
 
 static void
-fbCompositeSrcAdd_8888x8x8 (pixman_op_t op,
+fbCompositeSrcAdd_8888x8x8 (pixman_implementation_t *imp,
+			    pixman_op_t op,
 			    pixman_image_t * pSrc,
 			    pixman_image_t * pMask,
 			    pixman_image_t * pDst,
-			    int16_t      xSrc,
-			    int16_t      ySrc,
-			    int16_t      xMask,
-			    int16_t      yMask,
-			    int16_t      xDst,
-			    int16_t      yDst,
-			    uint16_t     width,
-			    uint16_t     height)
+			    int32_t      xSrc,
+			    int32_t      ySrc,
+			    int32_t      xMask,
+			    int32_t      yMask,
+			    int32_t      xDst,
+			    int32_t      yDst,
+			    int32_t     width,
+			    int32_t     height)
 {
     uint8_t	*dstLine, *dst;
     uint8_t	*maskLine, *mask;
@@ -943,18 +958,19 @@ fbCompositeSrcAdd_8888x8x8 (pixman_op_t op,
  */
 
 static void
-fbCompositeSolidFill (pixman_op_t op,
+fbCompositeSolidFill (pixman_implementation_t *imp,
+		      pixman_op_t op,
 		      pixman_image_t * pSrc,
 		      pixman_image_t * pMask,
 		      pixman_image_t * pDst,
-		      int16_t      xSrc,
-		      int16_t      ySrc,
-		      int16_t      xMask,
-		      int16_t      yMask,
-		      int16_t      xDst,
-		      int16_t      yDst,
-		      uint16_t     width,
-		      uint16_t     height)
+		      int32_t      xSrc,
+		      int32_t      ySrc,
+		      int32_t      xMask,
+		      int32_t      yMask,
+		      int32_t      xDst,
+		      int32_t      yDst,
+		      int32_t     width,
+		      int32_t     height)
 {
     uint32_t	src;
 
@@ -974,18 +990,19 @@ fbCompositeSolidFill (pixman_op_t op,
 }
 
 static void
-fbCompositeSrc_8888xx888 (pixman_op_t op,
+fbCompositeSrc_8888xx888 (pixman_implementation_t *imp,
+			  pixman_op_t op,
 			  pixman_image_t * pSrc,
 			  pixman_image_t * pMask,
 			  pixman_image_t * pDst,
-			  int16_t      xSrc,
-			  int16_t      ySrc,
-			  int16_t      xMask,
-			  int16_t      yMask,
-			  int16_t      xDst,
-			  int16_t      yDst,
-			  uint16_t     width,
-			  uint16_t     height)
+			  int32_t      xSrc,
+			  int32_t      ySrc,
+			  int32_t      xMask,
+			  int32_t      yMask,
+			  int32_t      xDst,
+			  int32_t      yDst,
+			  int32_t     width,
+			  int32_t     height)
 {
     uint32_t	*dst;
     uint32_t    *src;
@@ -1004,7 +1021,7 @@ fbCompositeSrc_8888xx888 (pixman_op_t op,
     }
 }
 
-static const FastPathInfo c_fast_path_array[] =
+static const FastPathInfo c_fast_paths[] =
 {
     { PIXMAN_OP_OVER, PIXMAN_solid,    PIXMAN_a8,       PIXMAN_r5g6b5,   fbCompositeSolidMask_nx8x0565, 0 },
     { PIXMAN_OP_OVER, PIXMAN_solid,    PIXMAN_a8,       PIXMAN_b5g6r5,   fbCompositeSolidMask_nx8x0565, 0 },
@@ -1053,4 +1070,281 @@ static const FastPathInfo c_fast_path_array[] =
     { PIXMAN_OP_NONE },
 };
 
-const FastPathInfo *const c_fast_paths = c_fast_path_array;
+static void
+fbCompositeSrcScaleNearest (pixman_implementation_t *imp,
+			    pixman_op_t     op,
+			    pixman_image_t *pSrc,
+			    pixman_image_t *pMask,
+			    pixman_image_t *pDst,
+			    int32_t         xSrc,
+			    int32_t         ySrc,
+			    int32_t         xMask,
+			    int32_t         yMask,
+			    int32_t         xDst,
+			    int32_t         yDst,
+			    int32_t        width,
+			    int32_t        height)
+{
+    uint32_t       *dst;
+    uint32_t       *src;
+    int             dstStride, srcStride;
+    int             i, j;
+    pixman_vector_t v;
+    
+    fbComposeGetStart (pDst, xDst, yDst, uint32_t, dstStride, dst, 1);
+    /* pass in 0 instead of xSrc and ySrc because xSrc and ySrc need to be
+     * transformed from destination space to source space */
+    fbComposeGetStart (pSrc, 0, 0, uint32_t, srcStride, src, 1);
+    
+    /* reference point is the center of the pixel */
+    v.vector[0] = pixman_int_to_fixed(xSrc) + pixman_fixed_1 / 2;
+    v.vector[1] = pixman_int_to_fixed(ySrc) + pixman_fixed_1 / 2;
+    v.vector[2] = pixman_fixed_1;
+    
+    if (!pixman_transform_point_3d (pSrc->common.transform, &v))
+        return;
+    
+    /* Round down to closest integer, ensuring that 0.5 rounds to 0, not 1 */
+    v.vector[0] -= pixman_fixed_e;
+    v.vector[1] -= pixman_fixed_e;
+    
+    for (j = 0; j < height; j++) {
+        pixman_fixed_t vx = v.vector[0];
+        pixman_fixed_t vy = v.vector[1];
+        for (i = 0; i < width; ++i) {
+            pixman_bool_t inside_bounds;
+            uint32_t result;
+            int x, y;
+            x = vx >> 16;
+            y = vy >> 16;
+	    
+            /* apply the repeat function */
+            switch (pSrc->common.repeat) {
+	    case PIXMAN_REPEAT_NORMAL:
+		x = MOD (x, pSrc->bits.width);
+		y = MOD (y, pSrc->bits.height);
+		inside_bounds = TRUE;
+		break;
+		
+	    case PIXMAN_REPEAT_PAD:
+		x = CLIP (x, 0, pSrc->bits.width-1);
+		y = CLIP (y, 0, pSrc->bits.height-1);
+		inside_bounds = TRUE;
+		break;
+		
+	    case PIXMAN_REPEAT_REFLECT:
+		x = MOD (x, pSrc->bits.width * 2);
+		if (x >= pSrc->bits.width)
+		    x = pSrc->bits.width * 2 - x - 1;
+		y = MOD (y, pSrc->bits.height * 2);
+		if (y >= pSrc->bits.height)
+		    y = pSrc->bits.height * 2 - y - 1;
+		inside_bounds = TRUE;
+		break;
+		
+	    case PIXMAN_REPEAT_NONE:
+	    default:
+		inside_bounds = (x >= 0 && x < pSrc->bits.width && y >= 0 && y < pSrc->bits.height);
+		break;
+            }
+	    
+            if (inside_bounds) {
+                //XXX: we should move this multiplication out of the loop
+                result = READ(pSrc, src + y * srcStride + x);
+            } else {
+                result = 0;
+            }
+            WRITE(pDst, dst + i, result);
+	    
+            /* adjust the x location by a unit vector in the x direction:
+             * this is equivalent to transforming x+1 of the destination point to source space */
+            vx += pSrc->common.transform->matrix[0][0];
+        }
+        /* adjust the y location by a unit vector in the y direction
+         * this is equivalent to transforming y+1 of the destination point to source space */
+        v.vector[1] += pSrc->common.transform->matrix[1][1];
+        dst += dstStride;
+    }
+}
+
+static void
+fast_path_composite (pixman_implementation_t *imp,
+		     pixman_op_t     op,
+		     pixman_image_t *src,
+		     pixman_image_t *mask,
+		     pixman_image_t *dest,
+		     int32_t         src_x,
+		     int32_t         src_y,
+		     int32_t         mask_x,
+		     int32_t         mask_y,
+		     int32_t         dest_x,
+		     int32_t         dest_y,
+		     int32_t        width,
+		     int32_t        height)
+{
+    if (src->type == BITS
+        && src->common.transform
+        && !mask
+        && op == PIXMAN_OP_SRC
+        && !src->common.alpha_map && !dest->common.alpha_map
+        && (src->common.filter == PIXMAN_FILTER_NEAREST)
+        && PIXMAN_FORMAT_BPP(dest->bits.format) == 32
+        && src->bits.format == dest->bits.format
+        && src->common.src_clip == &(src->common.full_region)
+        && !src->common.read_func && !src->common.write_func
+        && !dest->common.read_func && !dest->common.write_func)
+    {
+        /* ensure that the transform matrix only has a scale */
+        if (src->common.transform->matrix[0][1] == 0 &&
+            src->common.transform->matrix[1][0] == 0 &&
+            src->common.transform->matrix[2][0] == 0 &&
+            src->common.transform->matrix[2][1] == 0 &&
+            src->common.transform->matrix[2][2] == pixman_fixed_1)
+	{
+	    _pixman_walk_composite_region (imp, op,
+					   src, mask, dest,
+					   src_x, src_y,
+					   mask_x, mask_y,
+					   dest_x, dest_y,
+					   width, height,
+					   FALSE, FALSE, 
+					   fbCompositeSrcScaleNearest);
+	    return;
+	}
+    }
+
+    if (_pixman_run_fast_path (c_fast_paths, imp,
+			       op, src, mask, dest,
+			       src_x, src_y,
+			       mask_x, mask_y,
+			       dest_x, dest_y,
+			       width, height))
+    {
+	return;
+    }
+
+    _pixman_implementation_composite (imp->delegate, op,
+				      src, mask, dest,
+				      src_x, src_y,
+				      mask_x, mask_y,
+				      dest_x, dest_y,
+				      width, height);
+}
+
+static void
+pixman_fill8 (uint32_t  *bits,
+	      int	stride,
+	      int	x,
+	      int	y,
+	      int	width,
+	      int	height,
+	      uint32_t  xor)
+{
+    int byte_stride = stride * (int) sizeof (uint32_t);
+    uint8_t *dst = (uint8_t *) bits;
+    uint8_t v = xor & 0xff;
+    int i;
+
+    dst = dst + y * byte_stride + x;
+
+    while (height--)
+    {
+	for (i = 0; i < width; ++i)
+	    dst[i] = v;
+
+	dst += byte_stride;
+    }
+}
+
+static void
+pixman_fill16 (uint32_t *bits,
+	       int       stride,
+	       int       x,
+	       int       y,
+	       int       width,
+	       int       height,
+	       uint32_t  xor)
+{
+    int short_stride = (stride * (int) sizeof (uint32_t)) / (int) sizeof (uint16_t);
+    uint16_t *dst = (uint16_t *)bits;
+    uint16_t v = xor & 0xffff;
+    int i;
+
+    dst = dst + y * short_stride + x;
+
+    while (height--)
+    {
+	for (i = 0; i < width; ++i)
+	    dst[i] = v;
+
+	dst += short_stride;
+    }
+}
+
+static void
+pixman_fill32 (uint32_t *bits,
+	       int       stride,
+	       int       x,
+	       int       y,
+	       int       width,
+	       int       height,
+	       uint32_t  xor)
+{
+    int i;
+
+    bits = bits + y * stride + x;
+
+    while (height--)
+    {
+	for (i = 0; i < width; ++i)
+	    bits[i] = xor;
+
+	bits += stride;
+    }
+}
+
+static pixman_bool_t
+fast_path_fill (pixman_implementation_t *imp,
+		uint32_t *bits,
+		int stride,
+		int bpp,
+		int x,
+		int y,
+		int width,
+		int height,
+		uint32_t xor)
+{
+    switch (bpp)
+    {
+    case 8:
+	pixman_fill8 (bits, stride, x, y, width, height, xor);
+	break;
+	
+    case 16:
+	pixman_fill16 (bits, stride, x, y, width, height, xor);
+	break;
+	
+    case 32:
+	pixman_fill32 (bits, stride, x, y, width, height, xor);
+	break;
+	
+    default:
+	return _pixman_implementation_fill (
+	    imp->delegate, bits, stride, bpp, x, y, width, height, xor);
+	break;
+    }
+    
+    return TRUE;
+}
+
+pixman_implementation_t *
+_pixman_implementation_create_fast_path (void)
+{
+    pixman_implementation_t *general = _pixman_implementation_create_general ();
+    pixman_implementation_t *imp = _pixman_implementation_create (general);
+
+    imp->composite = fast_path_composite;
+    imp->fill = fast_path_fill;
+    
+    return imp;
+}
