@@ -95,10 +95,10 @@ static region_data_type_t *pixman_brokendata = (region_data_type_t *)&PREFIX(_br
  * the addresses of those structs which makes the existing code continue to
  * work.
  */
-PIXMAN_EXPORT void
-PREFIX(_set_static_pointers) (box_type_t *empty_box,
-				   region_data_type_t *empty_data,
-				   region_data_type_t *broken_data)
+void
+PREFIX(_internal_set_static_pointers) (box_type_t *empty_box,
+				       region_data_type_t *empty_data,
+				       region_data_type_t *broken_data)
 {
     pixman_region_emptyBox = empty_box;
     pixman_region_emptyData = empty_data;
@@ -572,7 +572,7 @@ pixman_region_appendNonO (
 {									\
     int newRects;							\
     if ((newRects = rEnd - r)) {					\
-	RECTALLOC(newReg, newRects);					\
+	RECTALLOC_BAIL(newReg, newRects, bail);					\
 	memmove((char *)PIXREGION_TOP(newReg),(char *)r, 			\
               newRects * sizeof(box_type_t));				\
 	newReg->data->numRects += newRects;				\
@@ -752,7 +752,8 @@ pixman_op(
 		bot = MIN(r1->y2, r2y1);
 		if (top != bot)	{
 		    curBand = newReg->data->numRects;
-		    pixman_region_appendNonO(newReg, r1, r1BandEnd, top, bot);
+		    if (!pixman_region_appendNonO(newReg, r1, r1BandEnd, top, bot))
+			goto bail;
 		    Coalesce(newReg, prevBand, curBand);
 		}
 	    }
@@ -763,7 +764,8 @@ pixman_op(
 		bot = MIN(r2->y2, r1y1);
 		if (top != bot) {
 		    curBand = newReg->data->numRects;
-		    pixman_region_appendNonO(newReg, r2, r2BandEnd, top, bot);
+		    if (!pixman_region_appendNonO(newReg, r2, r2BandEnd, top, bot))
+			goto bail;
 		    Coalesce(newReg, prevBand, curBand);
 		}
 	    }
@@ -779,8 +781,12 @@ pixman_op(
 	ybot = MIN(r1->y2, r2->y2);
 	if (ybot > ytop) {
 	    curBand = newReg->data->numRects;
-	    (* overlapFunc)(newReg, r1, r1BandEnd, r2, r2BandEnd, ytop, ybot,
-			    pOverlap);
+	    if (!(* overlapFunc)(newReg,
+				 r1, r1BandEnd,
+				 r2, r2BandEnd,
+				 ytop, ybot,
+				 pOverlap))
+		goto bail;
 	    Coalesce(newReg, prevBand, curBand);
 	}
 
@@ -805,7 +811,10 @@ pixman_op(
 	/* Do first nonOverlap1Func call, which may be able to coalesce */
 	FindBand(r1, r1BandEnd, r1End, r1y1);
 	curBand = newReg->data->numRects;
-	pixman_region_appendNonO(newReg, r1, r1BandEnd, MAX(r1y1, ybot), r1->y2);
+	if (!pixman_region_appendNonO(newReg,
+				      r1, r1BandEnd,
+				      MAX(r1y1, ybot), r1->y2))
+	    goto bail;
 	Coalesce(newReg, prevBand, curBand);
 	/* Just append the rest of the boxes  */
 	AppendRegions(newReg, r1BandEnd, r1End);
@@ -814,7 +823,10 @@ pixman_op(
 	/* Do first nonOverlap2Func call, which may be able to coalesce */
 	FindBand(r2, r2BandEnd, r2End, r2y1);
 	curBand = newReg->data->numRects;
-	pixman_region_appendNonO(newReg, r2, r2BandEnd, MAX(r2y1, ybot), r2->y2);
+	if (!pixman_region_appendNonO(newReg,
+				      r2, r2BandEnd,
+				      MAX(r2y1, ybot), r2->y2))
+	    goto bail;
 	Coalesce(newReg, prevBand, curBand);
 	/* Append rest of boxes */
 	AppendRegions(newReg, r2BandEnd, r2End);
@@ -840,6 +852,11 @@ pixman_op(
     }
 
     return TRUE;
+
+bail:
+    if (oldData)
+	free(oldData);
+    return pixman_break (newReg);
 }
 
 /*-
@@ -1219,109 +1236,6 @@ PREFIX(_union) (region_type_t *newReg,
  *	    Batch Rectangle Union
  *====================================================================*/
 
-/*-
- *-----------------------------------------------------------------------
- * pixman_region_append --
- *
- *      "Append" the rgn rectangles onto the end of dstrgn, maintaining
- *      knowledge of YX-banding when it's easy.  Otherwise, dstrgn just
- *      becomes a non-y-x-banded random collection of rectangles, and not
- *      yet a true region.  After a sequence of appends, the caller must
- *      call pixman_region_validate to ensure that a valid region is
- *      constructed.
- *
- * Results:
- *	TRUE if successful.
- *
- * Side Effects:
- *      dstrgn is modified if rgn has rectangles.
- *
- */
-PIXMAN_EXPORT pixman_bool_t
-PREFIX(_append) (region_type_t * dstrgn,
-		      region_type_t * rgn)
-{
-    int numRects, dnumRects, size;
-    box_type_t *new, *old;
-    int prepend;
-
-    if (PIXREGION_NAR(rgn))
-	return pixman_break (dstrgn);
-
-    if (!rgn->data && (dstrgn->data == pixman_region_emptyData))
-    {
-	dstrgn->extents = rgn->extents;
-	dstrgn->data = (region_data_type_t *)NULL;
-	return TRUE;
-    }
-
-    numRects = PIXREGION_NUM_RECTS(rgn);
-    if (!numRects)
-	return TRUE;
-    prepend = FALSE;
-    size = numRects;
-    dnumRects = PIXREGION_NUM_RECTS(dstrgn);
-    if (!dnumRects && (size < 200))
-	size = 200; /* XXX pick numbers out of a hat */
-    RECTALLOC(dstrgn, size);
-    old = PIXREGION_RECTS(rgn);
-    if (!dnumRects)
-	dstrgn->extents = rgn->extents;
-    else if (dstrgn->extents.x2 > dstrgn->extents.x1)
-    {
-	box_type_t *first, *last;
-
-	first = old;
-	last = PIXREGION_BOXPTR(dstrgn) + (dnumRects - 1);
-	if ((first->y1 > last->y2) ||
-	    ((first->y1 == last->y1) && (first->y2 == last->y2) &&
-	     (first->x1 > last->x2)))
-	{
-	    if (rgn->extents.x1 < dstrgn->extents.x1)
-		dstrgn->extents.x1 = rgn->extents.x1;
-	    if (rgn->extents.x2 > dstrgn->extents.x2)
-		dstrgn->extents.x2 = rgn->extents.x2;
-	    dstrgn->extents.y2 = rgn->extents.y2;
-	}
-	else
-	{
-	    first = PIXREGION_BOXPTR(dstrgn);
-	    last = old + (numRects - 1);
-	    if ((first->y1 > last->y2) ||
-		((first->y1 == last->y1) && (first->y2 == last->y2) &&
-		 (first->x1 > last->x2)))
-	    {
-		prepend = TRUE;
-		if (rgn->extents.x1 < dstrgn->extents.x1)
-		    dstrgn->extents.x1 = rgn->extents.x1;
-		if (rgn->extents.x2 > dstrgn->extents.x2)
-		    dstrgn->extents.x2 = rgn->extents.x2;
-		dstrgn->extents.y1 = rgn->extents.y1;
-	    }
-	    else
-		dstrgn->extents.x2 = dstrgn->extents.x1;
-	}
-    }
-    if (prepend)
-    {
-	new = PIXREGION_BOX(dstrgn, numRects);
-	if (dnumRects == 1)
-	    *new = *PIXREGION_BOXPTR(dstrgn);
-	else
-	    memmove((char *)new,(char *)PIXREGION_BOXPTR(dstrgn),
-		  dnumRects * sizeof(box_type_t));
-	new = PIXREGION_BOXPTR(dstrgn);
-    }
-    else
-	new = PIXREGION_BOXPTR(dstrgn) + dnumRects;
-    if (numRects == 1)
-	*new = *old;
-    else
-	memmove((char *)new, (char *)old, numRects * sizeof(box_type_t));
-    dstrgn->data->numRects += numRects;
-    return TRUE;
-}
-
 #define ExchangeRects(a, b) \
 {			    \
     box_type_t     t;	    \
@@ -1422,9 +1336,9 @@ QuickSortRects(
  *-----------------------------------------------------------------------
  */
 
-PIXMAN_EXPORT pixman_bool_t
-PREFIX(_validate) (region_type_t * badreg,
-		       int *pOverlap)
+static pixman_bool_t
+validate (region_type_t * badreg,
+	  int *pOverlap)
 {
     /* Descriptor for regions under construction  in Step 2. */
     typedef struct {
@@ -1433,6 +1347,8 @@ PREFIX(_validate) (region_type_t * badreg,
 	int	    curBand;
     } RegionInfo;
 
+    RegionInfo stack_regions[64];
+
 	     int	numRects;   /* Original numRects for badreg	    */
 	     RegionInfo *ri;	    /* Array of current regions		    */
     	     int	numRI;      /* Number of entries used in ri	    */
@@ -1482,10 +1398,8 @@ PREFIX(_validate) (region_type_t * badreg,
 
     /* Set up the first region to be the first rectangle in badreg */
     /* Note that step 2 code will never overflow the ri[0].reg rects array */
-    ri = (RegionInfo *) pixman_malloc_ab (4, sizeof(RegionInfo));
-    if (!ri)
-	return pixman_break (badreg);
-    sizeRI = 4;
+    ri = stack_regions;
+    sizeRI = sizeof (stack_regions) / sizeof (stack_regions[0]);
     numRI = 1;
     ri[0].prevBand = 0;
     ri[0].curBand = 0;
@@ -1554,9 +1468,16 @@ PREFIX(_validate) (region_type_t * badreg,
             data_size = sizeRI * sizeof(RegionInfo);
             if (data_size / sizeRI != sizeof(RegionInfo))
                 goto bail;
-            rit = (RegionInfo *) realloc(ri, data_size);
-	    if (!rit)
-		goto bail;
+	    if (ri == stack_regions) {
+		rit = malloc (data_size);
+		if (!rit)
+		    goto bail;
+		memcpy (rit, ri, numRI * sizeof (RegionInfo));
+	    } else {
+		rit = (RegionInfo *) realloc(ri, data_size);
+		if (!rit)
+		    goto bail;
+	    }
 	    ri = rit;
 	    rit = &ri[numRI];
 	}
@@ -1612,13 +1533,15 @@ NextRect: ;
 	    goto bail;
     }
     *badreg = ri[0].reg;
-    free(ri);
+    if (ri != stack_regions)
+	free(ri);
     good(badreg);
     return ret;
 bail:
     for (i = 0; i < numRI; i++)
 	freeData(&ri[i].reg);
-    free (ri);
+    if (ri != stack_regions)
+	free (ri);
 
     return pixman_break (badreg);
 }
@@ -2101,7 +2024,9 @@ PREFIX(_contains_point) (region_type_t * region,
         return(FALSE);
     if (numRects == 1)
     {
-	*box = region->extents;
+        if (box)
+	    *box = region->extents;
+
 	return(TRUE);
     }
     for (pbox = PIXREGION_BOXPTR(region), pboxEnd = pbox + numRects;
@@ -2114,7 +2039,10 @@ PREFIX(_contains_point) (region_type_t * region,
 	   break;		/* missed it */
 	if (x >= pbox->x2)
 	   continue;		/* not there yet */
-	*box = *pbox;
+
+        if (box)
+	    *box = *pbox;
+
 	return(TRUE);
     }
     return(FALSE);
@@ -2127,16 +2055,6 @@ PREFIX(_not_empty) (region_type_t * region)
     return(!PIXREGION_NIL(region));
 }
 
-PIXMAN_EXPORT void
-PREFIX(_empty) (region_type_t * region)
-{
-    good(region);
-    freeData(region);
-    region->extents.x2 = region->extents.x1;
-    region->extents.y2 = region->extents.y1;
-    region->data = pixman_region_emptyData;
-}
-
 PIXMAN_EXPORT box_type_t *
 PREFIX(_extents) (region_type_t * region)
 {
@@ -2233,5 +2151,5 @@ PREFIX(_init_rects) (region_type_t *region,
 
     /* Validate */
     region->extents.x1 = region->extents.x2 = 0;
-    return PREFIX(_validate) (region, &overlap);
+    return validate (region, &overlap);
 }
