@@ -28,164 +28,6 @@
 #include <stdlib.h>
 
 #include "pixman-private.h"
-#include "pixman-mmx.h"
-#include "pixman-sse2.h"
-
-#if defined(USE_SSE2) && defined(__GNUC__) && !defined(__x86_64__) && !defined(__amd64__)
-__attribute__((__force_align_arg_pointer__))
-#endif
-PIXMAN_EXPORT pixman_bool_t
-pixman_blt (uint32_t *src_bits,
-	    uint32_t *dst_bits,
-	    int src_stride,
-	    int dst_stride,
-	    int src_bpp,
-	    int dst_bpp,
-	    int src_x, int src_y,
-	    int dst_x, int dst_y,
-	    int width, int height)
-{
-#ifdef USE_SSE2
-    if (pixman_have_sse2())
-    {
-	return pixmanBltsse2 (src_bits, dst_bits, src_stride, dst_stride, src_bpp, dst_bpp,
-			      src_x, src_y, dst_x, dst_y, width, height);
-    }
-    else
-#endif
-#ifdef USE_MMX
-    if (pixman_have_mmx())
-    {
-	return pixman_blt_mmx (src_bits, dst_bits, src_stride, dst_stride, src_bpp, dst_bpp,
-			       src_x, src_y, dst_x, dst_y, width, height);
-    }
-    else
-#endif
-	return FALSE;
-}
-
-static void
-pixman_fill8 (uint32_t  *bits,
-	      int	stride,
-	      int	x,
-	      int	y,
-	      int	width,
-	      int	height,
-	      uint32_t  xor)
-{
-    int byte_stride = stride * (int) sizeof (uint32_t);
-    uint8_t *dst = (uint8_t *) bits;
-    uint8_t v = xor & 0xff;
-    int i;
-
-    dst = dst + y * byte_stride + x;
-
-    while (height--)
-    {
-	for (i = 0; i < width; ++i)
-	    dst[i] = v;
-
-	dst += byte_stride;
-    }
-}
-
-static void
-pixman_fill16 (uint32_t *bits,
-	       int       stride,
-	       int       x,
-	       int       y,
-	       int       width,
-	       int       height,
-	       uint32_t  xor)
-{
-    int short_stride = (stride * (int) sizeof (uint32_t)) / (int) sizeof (uint16_t);
-    uint16_t *dst = (uint16_t *)bits;
-    uint16_t v = xor & 0xffff;
-    int i;
-
-    dst = dst + y * short_stride + x;
-
-    while (height--)
-    {
-	for (i = 0; i < width; ++i)
-	    dst[i] = v;
-
-	dst += short_stride;
-    }
-}
-
-static void
-pixman_fill32 (uint32_t *bits,
-	       int       stride,
-	       int       x,
-	       int       y,
-	       int       width,
-	       int       height,
-	       uint32_t  xor)
-{
-    int i;
-
-    bits = bits + y * stride + x;
-
-    while (height--)
-    {
-	for (i = 0; i < width; ++i)
-	    bits[i] = xor;
-
-	bits += stride;
-    }
-}
-
-#if defined(USE_SSE2) && defined(__GNUC__) && !defined(__x86_64__) && !defined(__amd64__)
-__attribute__((__force_align_arg_pointer__))
-#endif
-PIXMAN_EXPORT pixman_bool_t
-pixman_fill (uint32_t *bits,
-	     int stride,
-	     int bpp,
-	     int x,
-	     int y,
-	     int width,
-	     int height,
-	     uint32_t xor)
-{
-#if 0
-    printf ("filling: %d %d %d %d (stride: %d, bpp: %d)   pixel: %x\n",
-	    x, y, width, height, stride, bpp, xor);
-#endif
-
-#ifdef USE_SSE2
-    if (pixman_have_sse2() && pixmanFillsse2 (bits, stride, bpp, x, y, width, height, xor))
-	return TRUE;
-#endif
-
-#ifdef USE_MMX
-    if (pixman_have_mmx() && pixman_fill_mmx (bits, stride, bpp, x, y, width, height, xor))
-	return TRUE;
-#endif
-    
-    switch (bpp)
-    {
-    case 8:
-	pixman_fill8 (bits, stride, x, y, width, height, xor);
-	break;
-	
-    case 16:
-	pixman_fill16 (bits, stride, x, y, width, height, xor);
-	break;
-	
-    case 32:
-	pixman_fill32 (bits, stride, x, y, width, height, xor);
-	break;
-	
-    default:
-	return FALSE;
-	break;
-    }
-
-    return TRUE;
-}
-
 
 /*
  * Compute the smallest value no less than y which is on a
@@ -432,6 +274,97 @@ pixman_version (void)
     return PIXMAN_VERSION;
 }
 
+/*
+ * Helper routine to expand a color component from 0 < n <= 8 bits to 16 bits by
+ * replication.
+ */
+static inline uint64_t
+expand16(const uint8_t val, int nbits)
+{
+    // Start out with the high bit of val in the high bit of result.
+    uint16_t result = (uint16_t)val << (16 - nbits);
+
+    if (nbits == 0)
+        return 0;
+
+    // Copy the bits in result, doubling the number of bits each time, until we
+    // fill all 16 bits.
+    while (nbits < 16) {
+        result |= result >> nbits;
+        nbits *= 2;
+    }
+
+    return result;
+}
+
+/*
+ * This function expands images from ARGB8 format to ARGB16.  To preserve
+ * precision, it needs to know the original source format.  For example, if the
+ * source was PIXMAN_x1r5g5b5 and the red component contained bits 12345, then
+ * the expanded value is 12345123.  To correctly expand this to 16 bits, it
+ * should be 1234512345123451 and not 1234512312345123.
+ */
+void
+pixman_expand(uint64_t *dst, const uint32_t *src,
+	      pixman_format_code_t format, int width)
+{
+    /*
+     * Determine the sizes of each component and the masks and shifts required
+     * to extract them from the source pixel.
+     */
+    const int a_size = PIXMAN_FORMAT_A(format),
+              r_size = PIXMAN_FORMAT_R(format),
+              g_size = PIXMAN_FORMAT_G(format),
+              b_size = PIXMAN_FORMAT_B(format);
+    const int a_shift = 32 - a_size,
+              r_shift = 24 - r_size,
+              g_shift = 16 - g_size,
+              b_shift =  8 - b_size;
+    const uint8_t a_mask = ~(~0 << a_size),
+                  r_mask = ~(~0 << r_size),
+                  g_mask = ~(~0 << g_size),
+                  b_mask = ~(~0 << b_size);
+    int i;
+
+    /* Start at the end so that we can do the expansion in place when src == dst */
+    for (i = width - 1; i >= 0; i--)
+    {
+        const uint32_t pixel = src[i];
+        // Extract the components.
+        const uint8_t a = (pixel >> a_shift) & a_mask,
+                      r = (pixel >> r_shift) & r_mask,
+                      g = (pixel >> g_shift) & g_mask,
+                      b = (pixel >> b_shift) & b_mask;
+        const uint64_t a16 = a_size ? expand16(a, a_size) : 0xffff,
+                       r16 = expand16(r, r_size),
+                       g16 = expand16(g, g_size),
+                       b16 = expand16(b, b_size);
+
+        dst[i] = a16 << 48 | r16 << 32 | g16 << 16 | b16;
+    }
+}
+
+/*
+ * Contracting is easier than expanding.  We just need to truncate the
+ * components.
+ */
+void
+pixman_contract(uint32_t *dst, const uint64_t *src, int width)
+{
+    int i;
+
+    /* Start at the beginning so that we can do the contraction in place when
+     * src == dst */
+    for (i = 0; i < width; i++)
+    {
+        const uint8_t a = src[i] >> 56,
+                      r = src[i] >> 40,
+                      g = src[i] >> 24,
+                      b = src[i] >> 8;
+        dst[i] = a << 24 | r << 16 | g << 8 | b;
+    }
+}
+
 /**
  * pixman_version_string:
  *
@@ -593,3 +526,269 @@ pixman_format_supported_source (pixman_format_code_t format)
 	return FALSE;
     }
 }
+
+void
+_pixman_walk_composite_region (pixman_implementation_t *imp,
+			      pixman_op_t op,
+			      pixman_image_t * pSrc,
+			      pixman_image_t * pMask,
+			      pixman_image_t * pDst,
+			      int16_t xSrc,
+			      int16_t ySrc,
+			      int16_t xMask,
+			      int16_t yMask,
+			      int16_t xDst,
+			      int16_t yDst,
+			      uint16_t width,
+			      uint16_t height,
+			      pixman_bool_t srcRepeat,
+			      pixman_bool_t maskRepeat,
+			      pixman_composite_func_t compositeRect)
+{
+    int		    n;
+    const pixman_box32_t *pbox;
+    int		    w, h, w_this, h_this;
+    int		    x_msk, y_msk, x_src, y_src, x_dst, y_dst;
+    pixman_region32_t reg;
+    pixman_region32_t *region;
+
+    pixman_region32_init (&reg);
+    if (!pixman_compute_composite_region32 (&reg, pSrc, pMask, pDst,
+					    xSrc, ySrc, xMask, yMask, xDst, yDst, width, height))
+    {
+	return;
+    }
+
+    region = &reg;
+
+    pbox = pixman_region32_rectangles (region, &n);
+    while (n--)
+    {
+	h = pbox->y2 - pbox->y1;
+	y_src = pbox->y1 - yDst + ySrc;
+	y_msk = pbox->y1 - yDst + yMask;
+	y_dst = pbox->y1;
+	while (h)
+	{
+	    h_this = h;
+	    w = pbox->x2 - pbox->x1;
+	    x_src = pbox->x1 - xDst + xSrc;
+	    x_msk = pbox->x1 - xDst + xMask;
+	    x_dst = pbox->x1;
+	    if (maskRepeat)
+	    {
+		y_msk = MOD (y_msk, pMask->bits.height);
+		if (h_this > pMask->bits.height - y_msk)
+		    h_this = pMask->bits.height - y_msk;
+	    }
+	    if (srcRepeat)
+	    {
+		y_src = MOD (y_src, pSrc->bits.height);
+		if (h_this > pSrc->bits.height - y_src)
+		    h_this = pSrc->bits.height - y_src;
+	    }
+	    while (w)
+	    {
+		w_this = w;
+		if (maskRepeat)
+		{
+		    x_msk = MOD (x_msk, pMask->bits.width);
+		    if (w_this > pMask->bits.width - x_msk)
+			w_this = pMask->bits.width - x_msk;
+		}
+		if (srcRepeat)
+		{
+		    x_src = MOD (x_src, pSrc->bits.width);
+		    if (w_this > pSrc->bits.width - x_src)
+			w_this = pSrc->bits.width - x_src;
+		}
+		(*compositeRect) (imp,
+				  op, pSrc, pMask, pDst,
+				  x_src, y_src, x_msk, y_msk, x_dst, y_dst,
+				  w_this, h_this);
+		w -= w_this;
+		x_src += w_this;
+		x_msk += w_this;
+		x_dst += w_this;
+	    }
+	    h -= h_this;
+	    y_src += h_this;
+	    y_msk += h_this;
+	    y_dst += h_this;
+	}
+	pbox++;
+    }
+    pixman_region32_fini (&reg);
+}
+
+static pixman_bool_t
+mask_is_solid (pixman_image_t *mask)
+{
+    if (mask->type == SOLID)
+	return TRUE;
+
+    if (mask->type == BITS &&
+	mask->common.repeat == PIXMAN_REPEAT_NORMAL &&
+	mask->bits.width == 1 &&
+	mask->bits.height == 1)
+    {
+	return TRUE;
+    }
+
+    return FALSE;
+}
+
+static const FastPathInfo *
+get_fast_path (const FastPathInfo *fast_paths,
+	       pixman_op_t         op,
+	       pixman_image_t     *pSrc,
+	       pixman_image_t     *pMask,
+	       pixman_image_t     *pDst,
+	       pixman_bool_t       is_pixbuf)
+{
+    const FastPathInfo *info;
+
+    for (info = fast_paths; info->op != PIXMAN_OP_NONE; info++)
+    {
+	pixman_bool_t valid_src		= FALSE;
+	pixman_bool_t valid_mask	= FALSE;
+
+	if (info->op != op)
+	    continue;
+
+	if ((info->src_format == PIXMAN_solid && pixman_image_can_get_solid (pSrc))		||
+	    (pSrc->type == BITS && info->src_format == pSrc->bits.format))
+	{
+	    valid_src = TRUE;
+	}
+
+	if (!valid_src)
+	    continue;
+
+	if ((info->mask_format == PIXMAN_null && !pMask)			||
+	    (pMask && pMask->type == BITS && info->mask_format == pMask->bits.format))
+	{
+	    valid_mask = TRUE;
+
+	    if (info->flags & NEED_SOLID_MASK)
+	    {
+		if (!pMask || !mask_is_solid (pMask))
+		    valid_mask = FALSE;
+	    }
+
+	    if (info->flags & NEED_COMPONENT_ALPHA)
+	    {
+		if (!pMask || !pMask->common.component_alpha)
+		    valid_mask = FALSE;
+	    }
+	}
+
+	if (!valid_mask)
+	    continue;
+	
+	if (info->dest_format != pDst->bits.format)
+	    continue;
+
+	if ((info->flags & NEED_PIXBUF) && !is_pixbuf)
+	    continue;
+
+	return info;
+    }
+
+    return NULL;
+}
+
+pixman_bool_t
+_pixman_run_fast_path (const FastPathInfo *paths,
+		       pixman_implementation_t *imp,
+		       pixman_op_t op,
+		       pixman_image_t *src,
+		       pixman_image_t *mask,
+		       pixman_image_t *dest,
+		       int32_t src_x,
+		       int32_t src_y,
+		       int32_t mask_x,
+		       int32_t mask_y,
+		       int32_t dest_x,
+		       int32_t dest_y,
+		       int32_t width,
+		       int32_t height)
+{
+    pixman_composite_func_t func = NULL;
+    pixman_bool_t src_repeat = src->common.repeat == PIXMAN_REPEAT_NORMAL;
+    pixman_bool_t mask_repeat = mask && mask->common.repeat == PIXMAN_REPEAT_NORMAL;
+    
+    if ((src->type == BITS || pixman_image_can_get_solid (src)) &&
+	(!mask || mask->type == BITS)
+        && !src->common.transform && !(mask && mask->common.transform)
+        && !(mask && mask->common.alpha_map) && !src->common.alpha_map && !dest->common.alpha_map
+        && (src->common.filter != PIXMAN_FILTER_CONVOLUTION)
+        && (src->common.repeat != PIXMAN_REPEAT_PAD)
+        && (src->common.repeat != PIXMAN_REPEAT_REFLECT)
+        && (!mask || (mask->common.filter != PIXMAN_FILTER_CONVOLUTION &&
+		      mask->common.repeat != PIXMAN_REPEAT_PAD &&
+		      mask->common.repeat != PIXMAN_REPEAT_REFLECT))
+	&& !src->common.read_func && !src->common.write_func
+	&& !(mask && mask->common.read_func)
+	&& !(mask && mask->common.write_func)
+	&& !dest->common.read_func
+	&& !dest->common.write_func)
+    {
+	const FastPathInfo *info;	
+	pixman_bool_t pixbuf;
+
+	pixbuf =
+	    src && src->type == BITS		&&
+	    mask && mask->type == BITS		&&
+	    src->bits.bits == mask->bits.bits	&&
+	    src_x == mask_x			&&
+	    src_y == mask_y			&&
+	    !mask->common.component_alpha	&&
+	    !mask_repeat;
+	
+	info = get_fast_path (paths, op, src, mask, dest, pixbuf);
+
+	if (info)
+	{
+	    func = info->func;
+		
+	    if (info->src_format == PIXMAN_solid)
+		src_repeat = FALSE;
+
+	    if (info->mask_format == PIXMAN_solid || info->flags & NEED_SOLID_MASK)
+		mask_repeat = FALSE;
+
+	    if ((src_repeat			&&
+		 src->bits.width == 1		&&
+		 src->bits.height == 1)	||
+		(mask_repeat			&&
+		 mask->bits.width == 1		&&
+		 mask->bits.height == 1))
+	    {
+		/* If src or mask are repeating 1x1 images and src_repeat or
+		 * mask_repeat are still TRUE, it means the fast path we
+		 * selected does not actually handle repeating images.
+		 *
+		 * So rather than call the "fast path" with a zillion
+		 * 1x1 requests, we just fall back to the general code (which
+		 * does do something sensible with 1x1 repeating images).
+		 */
+		func = NULL;
+	    }
+	}
+    }
+
+    if (func)
+    {
+	_pixman_walk_composite_region (imp, op,
+				       src, mask, dest,
+				       src_x, src_y, mask_x, mask_y,
+				       dest_x, dest_y,
+				       width, height,
+				       src_repeat, mask_repeat,
+				       func);
+	return TRUE;
+    }
+    
+    return FALSE;
+}
